"""
Test data
"""
import unittest
from enum import Enum
from pathlib import Path
from time import sleep

from oldaplib.src.connection import Connection
from oldaplib.src.dtypes.languagein import LanguageIn
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.dtypes.rdfset import RdfSet
from oldaplib.src.enums.language import Language
from oldaplib.src.enums.propertyclassattr import PropClassAttr
from oldaplib.src.enums.resourceclassattr import ResClassAttribute
from oldaplib.src.enums.xsd_datatypes import XsdDatatypes
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorAlreadyExists
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.helpers.semantic_version import SemanticVersion
from oldaplib.src.project import Project
from oldaplib.src.propertyclass import PropClassAttrContainer, PropertyClass, OwlPropertyType
from oldaplib.src.resourceclass import ResourceClass
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_decimal import Xsd_decimal
from oldaplib.src.xsd.xsd_integer import Xsd_integer
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_string import Xsd_string


class Graph(Enum):
    ONTO = 'test:onto'
    SHACL = 'test:shacl'


def check_prop_empty(con: Connection, context: Context, graph: Graph, res: str, prop: str) -> bool:
    sparql = context.sparql_context
    if graph == Graph.SHACL:
        sparql += f"""
        SELECT ?p ?v ?pp ?oo
        FROM {graph.value}
        WHERE {{
            {res}Shape sh:property ?prop .
            ?prop sh:path {prop} .
            ?prop ?p ?v .
            OPTIONAL {{ ?v ?pp ?oo }}
        }}
        """
    else:
        sparql += f"""
        SELECT ?p ?v ?pp ?oo
        FROM {graph.value}
        WHERE {{
            {res} rdfs:subClassOf ?prop .
            ?prop owl:onProperty test:propB .
            ?prop ?p ?v .
            OPTIONAL {{ ?v ?pp ?oo }}
        }}
        """
    jsonres = con.query(sparql)
    res = QueryProcessor(context, jsonres)
    return len(res) == 0

def check_res_empty(con: Connection, context: Context, graph: Graph, res: str) -> bool:
    sparql = context.sparql_context
    if graph == Graph.SHACL:
        res += 'Shape'
    sparql += f"""
    SELECT ?p ?v ?pp ?oo
    FROM {graph.value}
    WHERE {{
        {res} ?p ?v .
        OPTIONAL {{ ?v ?pp ?oo }}
    }}
    """
    jsonres = con.query(sparql)
    res = QueryProcessor(context, jsonres)
    return len(res) == 0


def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path


class TestResourceClass(unittest.TestCase):
    _context: Context
    _connection: Connection

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)

        cls._context = Context(name="DEFAULT")
        cls._context['test'] = NamespaceIRI("http://oldap.org/test#")
        cls._context.use('test', 'dcterms')

        cls._connection = Connection(server='http://localhost:7200',
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     repo="oldap",
                                     context_name="DEFAULT")

        cls._connection.clear_graph(Xsd_QName('test:shacl'))
        cls._connection.clear_graph(Xsd_QName('test:onto'))
        cls._connection.clear_graph(Xsd_QName('dcterms:shacl'))
        cls._connection.clear_graph(Xsd_QName('dcterms:onto'))

        file = project_root / 'oldaplib' / 'testdata' / 'connection_test.trig'
        cls._connection.upload_turtle(file)
        sleep(1)  # upload may take a while...
        cls._project = Project.read(cls._connection, "test")

    @classmethod
    def tearDownClass(cls):
        #cls._connection.clear_graph(QName('test:shacl'))
        #cls._connection.clear_graph(QName('test:onto'))
        pass

    # @unittest.skip('Work in progress')
    def test_constructor(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:testprop'),
                           subPropertyOf=Iri('test:comment'),
                           datatype=XsdDatatypes.langString,
                           name=LangString(["Test property@en", "Testprädikat@de"]),
                           description=LangString("A property for testing...@en"),
                           maxCount=Xsd_integer(1),
                           uniqueLang=Xsd_boolean(True),
                           languageIn=LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT),
                           order=Xsd_decimal(5))

        p2 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:enumprop'),
                           datatype=XsdDatatypes.string,
                           name=LangString(["Test enum@en", "Enumerationen@de"]),
                           maxCount=Xsd_integer(1),
                           minCount=Xsd_integer(1),
                           inSet=RdfSet(Xsd_string("yes"), Xsd_string("maybe"), Xsd_string("no")),
                           order=Xsd_decimal(6))

        properties: list[PropertyClass | Iri] = [
            Iri("test:comment"),
            Iri("test:test"),
            p1, p2
        ]

        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:TestResource"),
                           label=LangString(["Test resource@en", "Resource de test@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)
        self.assertEqual(r1[ResClassAttribute.LABEL], LangString(["Test resource@en", "Resource de test@fr"]))
        self.assertEqual(r1.label, LangString(["Test resource@en", "Resource de test@fr"]))
        self.assertEqual(r1[ResClassAttribute.COMMENT], LangString("For testing purposes@en"))
        self.assertEqual(r1.comment, LangString("For testing purposes@en"))
        self.assertTrue(r1[ResClassAttribute.CLOSED])
        self.assertTrue(r1.closed)

        prop1 = r1[Iri("test:comment")]
        self.assertIsNotNone(prop1)
        self.assertIsNone(prop1.internal)
        self.assertEqual(prop1.property_class_iri, Iri("test:comment"))
        self.assertEqual(prop1.datatype, XsdDatatypes.langString)
        self.assertEqual(prop1.name, LangString(["comment@en", "Kommentar@de"]))
        self.assertEqual(prop1.description, LangString("This is a test property@de"))
        self.assertEqual(prop1.maxCount, Xsd_integer(1))
        self.assertEqual(prop1.uniqueLang, Xsd_boolean(True))

        prop2 = r1[Iri("test:test")]
        self.assertIsNone(prop2.internal)
        self.assertEqual(prop2.property_class_iri, Iri("test:test"))
        self.assertEqual(prop2.datatype, XsdDatatypes.string)
        self.assertEqual(prop2.description, LangString("Property shape for testing purposes"))
        self.assertEqual(prop2.minCount, Xsd_integer(1))
        self.assertEqual(prop2.order, Xsd_decimal(3))

        prop3 = r1[Iri("test:testprop")]
        self.assertEqual(prop3.internal, Iri('test:TestResource'))
        self.assertEqual(prop3.property_class_iri, Iri("test:testprop"))
        self.assertEqual(prop3.propertyType, OwlPropertyType.OwlDataProperty)
        self.assertEqual(prop3.datatype, XsdDatatypes.langString)
        self.assertEqual(prop3.name, LangString(["Test property@en", "Testprädikat@de"]))
        self.assertEqual(prop3.order, Xsd_decimal(5))
        self.assertEqual(prop3.subPropertyOf, Iri("test:comment"))
        self.assertEqual(prop3.maxCount, Xsd_integer(1))
        self.assertEqual(prop3.uniqueLang,  Xsd_boolean(True))
        self.assertEqual(prop3.languageIn, LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT))

        prop4 = r1[Iri("test:enumprop")]
        self.assertEqual(prop4[PropClassAttr.IN],
                         RdfSet(Xsd_string("yes"), Xsd_string("maybe"), Xsd_string("no")))

    # @unittest.skip('Work in progress')
    def test_reading(self):
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:testMyRes'))
        self.assertEqual(r1.owl_class_iri, Iri('test:testMyRes'))
        self.assertEqual(r1.version, SemanticVersion(1, 0, 0))
        self.assertEqual(r1.creator, Iri('https://orcid.org/0000-0003-1681-4036'))
        self.assertEqual(r1.created, Xsd_dateTime('2023-11-04T12:00:00Z'))
        self.assertEqual(r1.contributor, Iri('https://orcid.org/0000-0003-1681-4036'))
        self.assertEqual(r1.modified, Xsd_dateTime('2023-11-04T12:00:00Z'))
        self.assertEqual(r1.label, LangString(["My Resource@en", "Meine Ressource@de", "Ma Resource@fr"]))
        self.assertEqual(r1.comment, LangString("Resource for testing..."))
        self.assertTrue(r1.closed)

        prop1 = r1[Iri('test:test')]
        self.assertIsNone(prop1.internal)
        self.assertEqual(prop1.property_class_iri, Iri("test:test"))
        self.assertEqual(prop1.version, SemanticVersion(1, 0, 0))
        self.assertEqual(prop1.creator, Iri('https://orcid.org/0000-0003-1681-4036'))
        self.assertEqual(prop1.created, Xsd_dateTime('2023-11-04T12:00:00Z'))
        self.assertEqual(prop1.contributor, Iri('https://orcid.org/0000-0003-1681-4036'))
        self.assertEqual(prop1.modified, Xsd_dateTime('2023-11-04T12:00:00Z'))
        self.assertEqual(prop1.propertyType, OwlPropertyType.OwlDataProperty)
        self.assertEqual(prop1.datatype, XsdDatatypes.string)
        self.assertEqual(prop1.description, LangString("Property shape for testing purposes"))
        self.assertEqual(prop1.minCount, Xsd_integer(1))
        self.assertEqual(prop1.order, Xsd_decimal(3))

        prop2 = r1[Iri('test:hasText')]
        self.assertEqual(prop2.internal, Iri('test:testMyRes'))
        self.assertEqual(prop2.property_class_iri, Iri("test:hasText"))
        self.assertEqual(prop2.version, SemanticVersion(1, 0, 0))
        self.assertEqual(prop2.creator, Iri('https://orcid.org/0000-0003-1681-4036'))
        self.assertEqual(prop2.created, Xsd_dateTime('2023-11-04T12:00:00Z'))
        self.assertEqual(prop2.contributor, Iri('https://orcid.org/0000-0003-1681-4036'))
        self.assertEqual(prop2.modified, Xsd_dateTime('2023-11-04T12:00:00Z'))
        self.assertEqual(prop2.propertyType, OwlPropertyType.OwlDataProperty)
        self.assertEqual(prop2.datatype, XsdDatatypes.langString)
        self.assertEqual(prop2.name, LangString(["A text@en", "Ein Text@de"]))
        self.assertEqual(prop2.description, LangString("A longer text..."))
        self.assertEqual(prop2.order, Xsd_decimal(1))
        self.assertEqual(prop2.minCount, Xsd_integer(1))
        self.assertEqual(prop2.maxCount, Xsd_integer(1))

        prop3 = r1[Iri('test:hasEnum')]
        self.assertEqual(prop3.propertyType, OwlPropertyType.OwlDataProperty)
        self.assertEqual(prop3.datatype, XsdDatatypes.string)
        self.assertEqual(prop3.inSet,
                         RdfSet(Xsd_string('red'), Xsd_string('green'), Xsd_string('blue'), Xsd_string('yellow')))

    def test_reading_with_superclass(self):
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:testMyResInherit'))
        self.assertEqual(r1.owl_class_iri, Iri('test:testMyResInherit'))
        self.assertEqual({Iri('test:testMyResMinimal')}, {x for x in r1.superclass})

    # @unittest.skip('Work in progress')
    def test_creating(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:testone'),
                           subPropertyOf=Iri('test:comment'),
                           datatype=XsdDatatypes.langString,
                           name=LangString(["Test property@en", "Testprädikat@de"]),
                           description=LangString("A property for testing...@en"),
                           minCount=Xsd_integer(1),
                           maxCount=Xsd_integer(1),
                           uniqueLang=Xsd_boolean(True),
                           languageIn=LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT),
                           order=Xsd_decimal(1))

        p2 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:testtwo'),
                           toNodeIri=Iri('test:testMyRes'),
                           name=LangString(["Excl. Test property@en", "Exkl. Testprädikat@de"]),
                           description=LangString("An exclusive property for testing...@en"),
                           minCount=Xsd_integer(1),
                           order=Xsd_decimal(2))

        p3 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:testthree'),
                           datatype=XsdDatatypes.int,
                           name=LangString(["E.N.U.M@en"]),
                           description=LangString("An exclusive enum testing...@en"),
                           inSet=RdfSet(Xsd_integer(1), Xsd_integer(2), Xsd_integer(3)),
                           order=Xsd_decimal(3))

        properties: list[PropertyClass | Iri] = [
            Iri("test:comment"),
            Iri("test:test"),
            p1, p2, p3
        ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:TestResource"),
                           label=LangString(["CreateResTest@en", "CréationResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)

        r1.create()

        r2 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri("test:TestResource"))
        self.assertEqual(r2.owl_class_iri, Xsd_QName("test:TestResource"))
        self.assertEqual(r2.label, LangString(["CreateResTest@en", "CréationResTeste@fr"]))
        self.assertEqual(r2.comment, LangString("For testing purposes@en"))
        self.assertTrue(r2.closed)

        prop1 = r2[Iri("test:comment")]
        self.assertIsNone(prop1.internal)
        self.assertEqual(prop1.property_class_iri, Iri('test:comment'))
        self.assertEqual(prop1.datatype, XsdDatatypes.langString)
        self.assertTrue(prop1.uniqueLang)
        self.assertEqual(prop1.maxCount, Xsd_integer(1))
        self.assertEqual(prop1.name, LangString(["comment@en", "Kommentar@de"]))
        self.assertEqual(prop1.description, LangString("This is a test property@de"))
        self.assertIsNone(prop1.subPropertyOf)
        self.assertEqual(prop1.order, Xsd_decimal(2))
        self.assertEqual(prop1.propertyType, OwlPropertyType.OwlDataProperty)
        self.assertEqual(prop1.creator, Iri('https://orcid.org/0000-0003-1681-4036'))
        self.assertEqual(prop1.created, Xsd_dateTime("2023-11-04T12:00:00Z"))

        prop2 = r2[Iri("test:test")]
        self.assertIsNone(prop1.internal)
        self.assertEqual(prop2.property_class_iri, Iri('test:test'))
        self.assertEqual(prop2.minCount, Xsd_integer(1))
        self.assertEqual(prop2.name, LangString("Test"))
        self.assertEqual(prop2.description, LangString("Property shape for testing purposes"))
        self.assertEqual(prop2.datatype, XsdDatatypes.string)
        self.assertEqual(prop2.order, Xsd_decimal(3))
        self.assertEqual(prop2.propertyType, OwlPropertyType.OwlDataProperty)

        prop3 = r2[Iri("test:testone")]
        self.assertEqual(prop3.internal, Iri("test:TestResource"))
        self.assertEqual(prop3.property_class_iri, Iri("test:testone"))
        self.assertEqual(prop3.datatype, XsdDatatypes.langString)
        self.assertEqual(prop3.name, LangString(["Test property@en", "Testprädikat@de"]))
        self.assertEqual(prop3.description, LangString("A property for testing...@en"))
        self.assertEqual(prop3.maxCount, Xsd_integer(1))
        self.assertEqual(prop3.minCount, Xsd_integer(1))
        self.assertEqual(prop3.languageIn, LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT))
        self.assertTrue(prop3.uniqueLang)
        self.assertEqual(prop3.order, Xsd_decimal(1))

        prop4 = r2[Iri("test:testtwo")]
        self.assertEqual(prop4.internal, Iri("test:TestResource"))
        self.assertEqual(prop4.property_class_iri, Iri("test:testtwo"))
        self.assertEqual(prop4.toNodeIri, Iri('test:testMyRes'))
        self.assertEqual(prop4.name, LangString(["Excl. Test property@en", "Exkl. Testprädikat@de"]))
        self.assertEqual(prop4.description, LangString("An exclusive property for testing...@en"))
        self.assertEqual(prop4.minCount, Xsd_integer(1))
        self.assertEqual(prop4.order, Xsd_decimal(2))

        prop5 = r2[Iri("test:testthree")]
        self.assertEqual(prop5.inSet, RdfSet(Xsd_integer(1), Xsd_integer(2), Xsd_integer(3)))

    def test_creating_with_superclass(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:sctest_prop1'),
                           datatype=XsdDatatypes.string,
                           name=LangString(["TestProp1"]),
                           order=Xsd_decimal(1))
        properties: list[PropertyClass | Iri] = [ p1 ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:ResWithSuperclasses"),
                           superclass={"test:testMyResMinimal", 'http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object'},
                           label=LangString(["CreateResTest@en", "CréationResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)
        r1.create()
        r2 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:ResWithSuperclasses'))
        s = set(r2.superclass.keys())
        self.assertEqual({"test:testMyResMinimal", 'http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object'}, s)

    def test_creating_with_thing_sc(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:thingtest_prop1'),
                           datatype=XsdDatatypes.string,
                           name=LangString(["ThingTestProp1"]),
                           order=Xsd_decimal(1))
        properties: list[PropertyClass | Iri] = [ p1 ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:ResWithSuperThing"),
                           superclass={'oldap:Thing', 'http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object'},
                           label=LangString(["CreateResTest@en", "CréationResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)
        r1.create()
        r2 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:ResWithSuperThing'))
        s = set(r2.superclass.keys())
        self.assertEqual({'oldap:Thing', Iri('http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object')}, s)

    # @unittest.skip('Work in progress')
    def test_double_creation(self):
        properties: list[PropertyClass | Iri] = [
            Iri("test:comment"),
            Iri("test:test"),
        ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:testMyResMinimal"),
                           label=LangString(["CreateResTest@en", "CréationResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)

        with self.assertRaises(OldapErrorAlreadyExists) as ex:
            r1.create()
        self.assertEqual(str(ex.exception), 'Object "test:testMyResMinimal" already exists.')

    # @unittest.skip('Work in progress')
    def test_updating_add(self):
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri("test:testMyResMinimal"))
        self.assertTrue(r1.closed)
        r1[ResClassAttribute.LABEL] = LangString(["Minimal Resource@en", "Kleinste Resource@de", "Plus petite ressource@fr"])
        r1[ResClassAttribute.COMMENT] = LangString("Eine Beschreibung einer minimalen Ressource")
        r1[ResClassAttribute.SUPERCLASS] = Iri("test:testMyRes")
        r1[ResClassAttribute.CLOSED] = Xsd_boolean(False)

        #
        # Add an external, shared property defined by its own sh:PropertyShape instance
        #
        r1[Iri('test:test')] = None

        #
        # Adding an internal, private property
        #
        p = PropertyClass(con=self._connection,
                          project=self._project,
                          toNodeIri=Iri('test:Person'),
                          maxCount=Xsd_integer(1))
        r1[Iri('dcterms:creator')] = p

        p2 = PropertyClass(con=self._connection,
                           project=self._project,
                           datatype=XsdDatatypes.string,
                           inSet=RdfSet(Xsd_string('A'), Xsd_string('B'), Xsd_string('C'), Xsd_string('D'))
                           )
        r1[Iri('test:color')] = p2
        r1.update()
        del r1
        r2 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri("test:testMyResMinimal"))
        self.assertEqual(LangString(["Minimal Resource@en", "Kleinste Resource@de", "Plus petite ressource@fr"]), r2.label)
        self.assertEqual(LangString("Eine Beschreibung einer minimalen Ressource"), r2.comment)
        self.assertEqual({Iri('test:testMyRes')}, set(r2.superclass))
        self.assertIsInstance(r2.superclass[Iri('test:testMyRes')], ResourceClass)
        self.assertFalse(r2[ResClassAttribute.CLOSED])

        prop1 = r2[Iri('test:test')]
        self.assertIsNone(prop1.internal)
        self.assertEqual(prop1.property_class_iri, Iri('test:test'))
        self.assertEqual(prop1.minCount, Xsd_integer(1))
        self.assertEqual(prop1.name, LangString("Test"))
        self.assertEqual(prop1.description, LangString("Property shape for testing purposes"))
        self.assertEqual(prop1.datatype, XsdDatatypes.string)
        self.assertEqual(prop1.order, Xsd_decimal(3))
        self.assertEqual(prop1.propertyType, OwlPropertyType.OwlDataProperty)

        prop2 = r2[Iri('dcterms:creator')]
        self.assertEqual(prop2.internal, Xsd_QName("test:testMyResMinimal"))
        self.assertEqual(prop2.toNodeIri, Xsd_QName('test:Person'))
        self.assertEqual(prop2.maxCount, Xsd_integer(1))
        self.assertEqual(prop1.propertyType, OwlPropertyType.OwlDataProperty)

        prop3 = r2[Iri('test:color')]
        self.assertEqual(prop3.internal, Xsd_QName("test:testMyResMinimal"))
        self.assertEqual(prop3.inSet, RdfSet(Xsd_string('A'), Xsd_string('B'), Xsd_string('C'), Xsd_string('D')))

        sparql = self._context.sparql_context
        sparql += """
        SELECT ?p ?v
        FROM test:onto
        WHERE {
            test:testMyResMinimal rdfs:subClassOf ?prop .
            ?prop owl:onProperty dcterms:creator .
            ?prop ?p ?v .
        }
        """
        jsonobj = self._connection.query(sparql)
        res = QueryProcessor(self._context, jsonobj)
        result = {
            Iri('rdf:type'): 'owl:Restriction',
            Iri('owl:onProperty'): 'dcterms:creator',
            Iri('owl:maxQualifiedCardinality'): 1,
            Iri('owl:onClass'): 'test:Person'
        }
        for r in res:
            p = r['p']
            v = r['v']
            self.assertEqual(result[p], v)

        sparql = self._context.sparql_context
        sparql += """
        SELECT ?p ?v
        FROM test:onto
        WHERE {
            test:testMyResMinimal rdfs:subClassOf ?prop .
            ?prop owl:onProperty test:test .
            ?prop ?p ?v .
        }
        """
        jsonobj = self._connection.query(sparql)
        res = QueryProcessor(self._context, jsonobj)
        result = {
            Iri('rdf:type'): Iri('owl:Restriction'),
            Iri('owl:onProperty'): Iri('test:test'),
            Iri('owl:minQualifiedCardinality'): Xsd_integer(1),
            Iri('owl:onDatatype'): Iri("xsd:string"),
        }
        for r in res:
            p = r['p']
            v = r['v']
            self.assertEqual(result[p], v)

    # @unittest.skip('Work in progress')
    def test_updating(self):
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri("test:testMyRes"))
        self.assertEqual(r1[Iri('test:hasText')].maxCount, Xsd_integer(1))
        self.assertEqual(r1[Iri('test:hasText')].minCount, Xsd_integer(1))
        self.assertEqual(r1[Iri('test:hasText')].languageIn, LanguageIn(Language.EN, Language.DE))
        r1.label[Language.IT] = "La mia risorsa"
        r1.closed = Xsd_boolean(False)
        r1[ResClassAttribute.SUPERCLASS] = {Iri('dcterms:TopGaga')}
        r1[Iri('test:hasText')].name[Language.FR] = "Un Texte Français"
        r1[Iri('test:hasText')].maxCount = Xsd_integer(12)
        r1[Iri('test:hasText')].languageIn = LanguageIn(Language.DE, Language.FR, Language.IT)
        r1[Iri('test:hasEnum')].inSet = RdfSet(Xsd_string('L'), Xsd_string('a'), Xsd_string('b'))
        r1.update()

        r2 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri("test:testMyRes"))
        self.assertEqual(r2.label, LangString(["My Resource@en", "Meine Ressource@de", "Ma Resource@fr", "La mia risorsa@it"]))
        self.assertFalse(r2.closed)
        self.assertEqual({Iri('dcterms:TopGaga')}, set(r2.superclass))
        self.assertIsNone(r2.superclass[Iri('dcterms:TopGaga')])
        self.assertEqual(r2[Iri('test:hasText')].name, LangString(["A text@en", "Ein Text@de", "Un Texte Français@fr"]))
        self.assertEqual(r2[Iri('test:hasText')].maxCount, Xsd_integer(12))
        self.assertEqual(r2[Iri('test:hasText')].languageIn, LanguageIn(Language.DE, Language.FR, Language.IT))
        self.assertEqual(r2[Iri('test:hasEnum')].inSet, RdfSet(Xsd_string('L'), Xsd_string('a'), Xsd_string('b')))

    def test_updating_sc_A(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:p1'),
                           datatype=XsdDatatypes.string,
                           name=LangString(["P1"]),
                           order=Xsd_decimal(1))
        properties: list[PropertyClass | Iri] = [ p1 ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:Crazy"),
                           superclass={"oldap:Thing", 'http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object'},
                           label=LangString(["CreateResTest@en", "CréationResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)
        r1.create()
        del r1
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:Crazy'))
        del r1.superclass[Iri('http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object')]
        r1.update()
        del r1
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:Crazy'))
        self.assertEqual({"oldap:Thing"}, set(r1.superclass))

    def test_updating_sc_B(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:p2'),
                           datatype=XsdDatatypes.string,
                           name=LangString(["P2"]),
                           order=Xsd_decimal(1))
        properties: list[PropertyClass | Iri] = [ p1 ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:CrazyB"),
                           superclass={"test:testMyRes", 'http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object'},
                           label=LangString(["CreateResTest@en", "CréationResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)
        r1.create()
        del r1
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:CrazyB'))
        del r1.superclass[Iri("test:testMyRes")]
        r1.update()
        del r1
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:CrazyB'))
        self.assertEqual({Iri('http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object')}, set(r1.superclass))

    def test_updateing_sc_C(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:p3'),
                           datatype=XsdDatatypes.string,
                           name=LangString(["P3"]),
                           order=Xsd_decimal(1))
        properties: list[PropertyClass | Iri] = [ p1 ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:CrazyC"),
                           superclass={"test:testMyRes", 'http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object', 'test:testMyResMinimal'},
                           label=LangString(["CreateResTest@en", "CréationResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)
        r1.create()
        del r1
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:CrazyC'))
        del r1.superclass['test:testMyResMinimal']
        r1.superclass[Iri('http://gugus.com/gaga/wird/nicht/gehen')] = None
        r1.update()
        r1 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri('test:CrazyC'))
        self.assertEqual({"test:testMyRes",
                          'http://www.cidoc-crm.org/cidoc-crm/E22_Man-Made_Object',
                          'http://gugus.com/gaga/wird/nicht/gehen'}, set(r1.superclass))

    # @unittest.skip('Work in progress')
    def test_delete_props(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:propA'),
                           subPropertyOf=Iri('test:comment'),
                           datatype=XsdDatatypes.langString,
                           name=LangString(["Test property@en", "Testprädikat@de"]),
                           description=LangString("A property for testing...@en"),
                           maxCount=Xsd_integer(1),
                           minCount=Xsd_integer(1),
                           uniqueLang=Xsd_boolean(True),
                           languageIn=LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT),
                           order=Xsd_decimal(1))

        p2 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:propB'),
                           toNodeIri=Iri('test:testMyRes'),
                           name=LangString(["Excl. Test property@en", "Exkl. Testprädikat@de"]),
                           description=LangString("An exclusive property for testing...@en"),
                           minCount=Xsd_integer(1),
                           order=Xsd_decimal(2))

        p3 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:propC'),
                           datatype=XsdDatatypes.int,
                           inSet=RdfSet(Xsd_integer(10), Xsd_integer(20), Xsd_integer(30)))

        properties: list[PropertyClass | Iri] = [
            Iri("test:comment"),
            Iri("test:test"),
            p1, p2, p3
        ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:TestResourceDelProps"),
                           superclass="test:testMyResMinimal",
                           label=LangString(["CreateResTest@en", "CréationResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)

        r1.create()

        r2 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri("test:TestResourceDelProps"))
        del r2[Iri('test:propB')]
        del r2[Iri("test:test")]  # OWL is not yet removed (rdfs:subClassOf is still there)
        del r2[Iri('test:propC')]
        r2.update()

        r3 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri("test:TestResourceDelProps"))

        self.assertTrue(check_prop_empty(self._connection, self._context, Graph.SHACL, 'test:testMyResMinimal', 'test:propB'))
        self.assertTrue(check_prop_empty(self._connection, self._context, Graph.ONTO, 'test:testMyResMinimal', 'test:propB'))

        self.assertTrue(check_prop_empty(self._connection, self._context, Graph.SHACL, 'test:testMyResMinimal', 'test:test'))
        self.assertTrue(check_prop_empty(self._connection, self._context, Graph.ONTO, 'test:testMyResMinimal', 'test:test'))

        self.assertTrue(check_prop_empty(self._connection, self._context, Graph.SHACL, 'test:testMyResMinimal', 'test:propC'))
        self.assertTrue(check_prop_empty(self._connection, self._context, Graph.ONTO, 'test:testMyResMinimal', 'test:propC'))

    # @unittest.skip('Work in progress')
    def test_delete(self):
        p1 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:deleteA'),
                           subPropertyOf=Iri('test:comment'),
                           datatype=XsdDatatypes.langString,
                           name=LangString(["Test property@en", "Testprädikat@de"]),
                           description=LangString("A property for testing...@en"),
                           maxCount=Xsd_integer(1),
                           minCount=Xsd_integer(1),
                           uniqueLang=Xsd_boolean(True),
                           languageIn=LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT),
                           order=Xsd_decimal(1))

        p2 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:deleteB'),
                           toNodeIri=Iri('test:testMyRes'),
                           name=LangString(["Excl. Test property@en", "Exkl. Testprädikat@de"]),
                           description=LangString("A property for testing...@en"),
                           minCount=Xsd_integer(1),
                           order=Xsd_decimal(2))

        p3 = PropertyClass(con=self._connection,
                           project=self._project,
                           property_class_iri=Iri('test:deleteC'),
                           datatype=XsdDatatypes.int,
                           inSet=RdfSet(Xsd_integer(10), Xsd_integer(20), Xsd_integer(30)))

        properties: list[PropertyClass | Iri] = [
            Iri("test:comment"),
            Iri("test:test"),
            p1, p2, p3
        ]
        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:TestResourceDelete"),
                           superclass="test:testMyResMinimal",
                           label=LangString(["DeleteResTest@en", "EffaçerResTeste@fr"]),
                           comment=LangString("For testing purposes@en"),
                           closed=Xsd_boolean(True),
                           properties=properties)
        r1.create()
        del r1

        r2 = ResourceClass.read(con=self._connection,
                                project=self._project,
                                owl_class_iri=Iri("test:TestResourceDelete"))
        r2.delete()

        self.assertTrue(check_res_empty(self._connection, self._context, Graph.SHACL, 'test:TestResourceDelete'))
        self.assertTrue(check_res_empty(self._connection, self._context, Graph.ONTO, 'test:TestResourceDelete'))
        superclass = ResourceClass.read(con=self._connection,
                                        project=self._project,
                                        owl_class_iri=Iri("test:testMyResMinimal"))
        self.assertEqual(Iri("test:testMyResMinimal"), superclass.owl_class_iri)

    # @unittest.skip('Work in progress')
    def test_write_trig(self):
        project_id = PropertyClass(con=self._connection,
                                   project=self._project,
                                   property_class_iri=Iri('test:projectId'),
                                   datatype=XsdDatatypes.langString,
                                   name=LangString(["Project ID@en", "Projekt ID@de"]),
                                   description=LangString(["Unique ID for project@en", "Eindeutige ID für Projekt@de"]),
                                   minCount=Xsd_integer(1),
                                   languageIn=LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT),
                                   uniqueLang=Xsd_boolean(True),
                                   order=Xsd_decimal(1))
        project_name = PropertyClass(con=self._connection,
                                     project=self._project,
                                     property_class_iri=Iri('test:projectName'),
                                     datatype=XsdDatatypes.langString,
                                     name=LangString(["Project name@en", "Projektname@de"]),
                                     description=LangString(["A description of the project@en", "EineBeschreibung des Projekts@de"]),
                                     minCount=Xsd_integer(1),
                                     languageIn=LanguageIn(Language.EN, Language.DE, Language.FR, Language.IT),
                                     uniqueLang=Xsd_boolean(True),
                                     order=Xsd_decimal(2))

        properties: list[PropertyClass | Xsd_QName] = [
            project_id, project_name
        ]
        superclass = ResourceClass.read(con=self._connection,
                                        project=self._project,
                                        owl_class_iri=Iri("test:testMyResMinimal"))

        r1 = ResourceClass(con=self._connection,
                           project=self._project,
                           owlclass_iri=Iri("test:Project"),
                           superclass={"test:testMyResMinimal", Iri('http://andromeda.com/cluster1/cepheid_42')},
                           label=LangString(["Project@en", "Projekt@de"]),
                           comment=LangString(["Definiton of a project@en", "Definition eines Projektes@de"]),
                           closed=Xsd_boolean(True),
                           properties=properties)
        r1.write_as_trig("gaga.trig")


if __name__ == '__main__':
    unittest.main()



import unittest

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.semantic_version import SemanticVersion


class TestSemanticVersion(unittest.TestCase):
    def test_constructor(self):
        sv = SemanticVersion()
        self.assertEqual(sv.major, 0)
        self.assertEqual(sv.minor, 1)
        self.assertEqual(sv.patch, 0)

        sv = SemanticVersion(1, 2,3)
        self.assertEqual(sv.major, 1)
        self.assertEqual(sv.minor, 2)
        self.assertEqual(sv.patch, 3)

        sv = SemanticVersion.fromString("1.2.3")
        self.assertEqual(sv.major, 1)
        self.assertEqual(sv.minor, 2)
        self.assertEqual(sv.patch, 3)

        with self.assertRaises(OldapErrorValue):
            sv = SemanticVersion.fromString("1.2.A")

    def test_string_conversion(self):
        sv = SemanticVersion(1, 2,3)
        self.assertEqual(str(sv), '1.2.3')

    def test_repr_conversion(self):
        sv = SemanticVersion(1, 2,3)
        self.assertEqual(repr(sv), 'SemanticVersion(1, 2, 3)')

    def test_rdf_conversion(self):
        sv = SemanticVersion(1, 2, 3)
        self.assertEqual(sv.toRdf, '"1.2.3"^^xsd:string')

    def test_compare_ops(self):
        self.assertTrue(SemanticVersion(1, 2, 3) == SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(1, 2, 3) >= SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(1, 2, 4) >= SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(1, 3, 3) >= SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(2, 3, 3) >= SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(1, 2, 3) <= SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(1, 2, 2) <= SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(1, 1, 3) <= SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(0, 2, 3) <= SemanticVersion(1, 2, 3))
        self.assertFalse(SemanticVersion(1, 2, 3) != SemanticVersion(1, 2, 3))

        self.assertTrue(SemanticVersion(1, 2, 3) != SemanticVersion(1, 2, 4))
        self.assertTrue(SemanticVersion(1, 2, 3) != SemanticVersion(1, 3, 3))
        self.assertTrue(SemanticVersion(1, 2, 3) != SemanticVersion(2, 2, 3))

        self.assertTrue(SemanticVersion(1, 2, 3) < SemanticVersion(1, 2, 4))
        self.assertTrue(SemanticVersion(1, 2, 3) < SemanticVersion(1, 3, 3))
        self.assertTrue(SemanticVersion(1, 2, 3) < SemanticVersion(2, 2, 3))

        self.assertTrue(SemanticVersion(1, 2, 4) > SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(1, 3, 3) > SemanticVersion(1, 2, 3))
        self.assertTrue(SemanticVersion(2, 2, 3) > SemanticVersion(1, 2, 3))
        self.assertFalse(SemanticVersion(2, 2, 3) > SemanticVersion(2, 2, 3))

    def test_increment(self):
        sv = SemanticVersion(1, 2, 3)
        sv.increment_patch()
        self.assertEqual(sv.major, 1)
        self.assertEqual(sv.minor, 2)
        self.assertEqual(sv.patch, 4)

        sv = SemanticVersion(1, 2, 3)
        sv.increment_minor()
        self.assertEqual(sv.major, 1)
        self.assertEqual(sv.minor, 3)
        self.assertEqual(sv.patch, 0)


        sv = SemanticVersion(1, 2, 3)
        sv.increment_major()
        self.assertEqual(sv.major, 2)
        self.assertEqual(sv.minor, 0)
        self.assertEqual(sv.patch, 0)


if __name__ == '__main__':
    unittest.main()



import re
from typing import Tuple, Self

from pystrict import strict

from oldaplib.src.enums.xsd_datatypes import XsdValidator, XsdDatatypes
from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_gYearMonth(Xsd):
    """
    Implementation of the XML Schema [xsd:gYearMonth](https://www.w3.org/TR/xmlschema11-2/#gYearMonth) datatype
    """
    __year: int
    __month: int
    __tz: Tuple[int, int] | None
    __zulu: bool

    def __init__(self, value: Self | str, validate: bool = True):
        """
        Constructor of the Xsd_gYearMonth class
        :param value: Xsd_gYearMonth instance or a valid string representing a Xsd_gYearMonth instance
        :type value: Xsd_gYearMonth | str
        :raises OldapErrorValue: if the value is not a valid Xsd_gYearMonth instance
        """
        if isinstance(value, Xsd_gYearMonth):
            self.__year = value.__year
            self.__month = value.__month
            self.__tz = value.__tz
        else:
            if validate:
                if not XsdValidator.validate(XsdDatatypes.gYearMonth, value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:gYearMonth')
                if not re.match("([+-]?[0-9]{4})-([0-9]{2})((([+-][0-9]{2}):([0-9]{2}))|(Z))?", value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:gYearMonth.')
            res = re.split("([+-]?[0-9]{4})-([0-9]{2})((([+-][0-9]{2}):([0-9]{2}))|(Z))?", value)
            if len(res) != 9:
                raise OldapErrorValue(f'Invalid string "{value}" for xsd:gYearMonth.')
            self.__year = int(res[1])
            self.__month = int(res[2])
            if res[3] == 'Z':
                self.__tz = (0, 0)
                self.__zulu = True
            elif res[3] is not None:
                self.__tz = (int(res[5]), int(res[6]))
                self.__zulu = False
            else:
                self.__tz = None
        if self.__month < 1 or self.__month > 12:
            raise OldapErrorValue(f'Invalid string "{value}" for xsd:gYearMonth.')

    def __str__(self):
        """
        String representation of the Xsd_gYearMonth instance
        :return: String representation of the Xsd_gYearMonth instance
        """
        ff = '05' if self.__year < 0 else '04'
        s = f'{self.__year:{ff}}-{self.__month:02}'
        if self.__tz is not None:
            if self.__zulu:
                s += 'Z'
            else:
                s += f'{self.__tz[0]:0=+3}:{self.__tz[1]:02}'
        return s

    def __repr__(self):
        """
        Constructor string representation of the Xsd_gYearMonth instance
        :return: Constructor string
        """
        return f'Xsd_gYearMonth("{str(self)}")'

    def __eq__(self, other: Self | str | None) -> bool:
        """
        Equality test for Xsd_gYearMonth instance
        :param other: Xsd_gYearMonth instance or a valid string representing a Xsd_gYearMonth instance
        :type other: Xsd_gYearMonth | str | None
        :return: True or False
        :raise OldapErrorValue: if the value is not a valid Xsd_gYearMonth instance
        """
        if other is None:
            return False
        if not isinstance(other, Xsd_gYearMonth):
            other = Xsd_gYearMonth(other)
        if self.__year != other.__year:
            return False
        if self.__month != other.__month:
            return False
        if self.__tz != other.__tz:
            return False
        return True

    def __hash__(self) -> int:
        """
        Hash representation of the Xsd_gYearMonth instance
        :return: Hash value
        """
        return hash(str(self))

    def _as_dict(self) -> dict[str, str]:
        """
        Internal method for JSON serialization (@serializer decorator)
        :return: dict
        """
        return {'value': str(self)}

    @property
    def toRdf(self) -> str:
        """
        RDF representation of the Xsd_gYearMonth instance
        :return: RDF string
        """
        return f'"{str(self)}"^^xsd:gYearMonth'




from datetime import datetime
from typing import List, Set, Dict, Tuple, Optional, Any, Union

from pystrict import strict

from oldaplib.src.helpers.context import Context
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorNotFound
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.helpers.tools import lprint
from oldaplib.src.iconnection import IConnection


#@strict
class Model:
    _con: IConnection
    _changed: Set[str]

    def __init__(self, connection: IConnection) -> None:
        # if not isinstance(connection, Connection):
        #     raise OldapError('"con"-parameter must be an instance of Connection')
        # if type(connection) != Connection:
        #     raise OldapError('"con"-parameter must be an instance of Connection')
        self._con = connection
        self._changed = set()

    def has_changed(self) -> bool:
        if self._changed:
            return True
        else:
            return False

    def get_modified_by_iri(self, graph: Xsd_QName, iri: Iri) -> Xsd_dateTime:
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        SELECT ?modified
        FROM {graph}
        WHERE {{
            {iri.toRdf} dcterms:modified ?modified
        }}
        """
        jsonobj = None
        if self._con.in_transaction():
            jsonobj = self._con.transaction_query(sparql)
        else:
            jsonobj = self._con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        if len(res) != 1:
            raise OldapErrorNotFound(f'No resource found with iri "{iri}".')
        for r in res:
            return r['modified']

    def set_modified_by_iri(self,
                            graph: Xsd_QName,
                            iri: Iri,
                            old_timestamp: Xsd_dateTime,
                            timestamp: Xsd_dateTime) -> None:
        try:
            context = Context(name=self._con.context_name)
            sparql = context.sparql_context
            sparql += f"""
            WITH {graph}
            DELETE {{
                ?res dcterms:modified {old_timestamp.toRdf} .
                ?res dcterms:contributor ?contributor .
            }}
            INSERT {{
                ?res dcterms:modified {timestamp.toRdf} .
                ?res dcterms:contributor {self._con.userIri.toRdf} .
            }}
            WHERE {{
                BIND({iri.toRdf} as ?res)
                ?res dcterms:modified {old_timestamp.toRdf} .
                ?res dcterms:contributor ?contributor .
            }}
            """
        except Exception as e:
            raise OldapError(e)
        if self._con.in_transaction():
            self._con.transaction_update(sparql)
        else:
            self._con.query(sparql)



from typing import Self

from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorValue


class SemanticVersion:
    __major: int
    __minor: int
    __patch: int

    def __init__(self, major: int = 0, minor: int = 1, patch: int = 0):
        self.__major = major
        self.__minor = minor
        self.__patch = patch

    def __str__(self) -> str:
        return f"{self.__major}.{self.__minor}.{self.__patch}"

    def __repr__(self) -> str:
        return f'SemanticVersion({self.__major}, {self.__minor}, {self.__patch})'

    def __eq__(self, other: Self) -> bool:
        return self.__major == other.__major and self.__minor == other.__minor and self.__patch == other.__patch

    def __ne__(self, other: Self) -> bool:
        return self.__major != other.__major or self.__minor != other.__minor or self.__patch != other.__patch

    def __gt__(self, other: Self) -> bool:
        if self.__major > other.__major:
            return True
        if self.__major == other.__major and self.__minor > other.__minor:
            return True
        if self.__major == other.__major and self.__minor == other.__minor and self.__patch > other.__patch:
            return True
        return False

    def __ge__(self, other: Self) -> bool:
        if self.__major >= other.__major:
            return True
        if self.__major == other.__major and self.__minor >= other.__minor:
            return True
        if self.__major == other.__major and self.__minor == other.__minor and self.__patch >= other.__patch:
            return True
        return False

    def __lt__(self, other: Self) -> bool:
        if self.__major < other.__major:
            return True
        if self.__major == other.__major and self.__minor < other.__minor:
            return True
        if self.__major == other.__major and self.__minor == other.__minor and self.__patch < other.__patch:
            return True
        return False

    def __le__(self, other: Self) -> bool:
        if self.__major <= other.__major:
            return True
        if self.__major == other.__major and self.__minor <= other.__minor:
            return True
        if self.__major == other.__major and self.__minor == other.__minor and self.__patch <= other.__patch:
            return True
        return False

    @property
    def toRdf(self) -> str:
        return f'"{self.__major}.{self.__minor}.{self.__patch}"^^xsd:string'

    @classmethod
    def fromString(cls, versionstring: str) -> Self:
        try:
            major, minor, patch = str(versionstring).split(".")
            return cls(int(major), int(minor), int(patch))
        except ValueError as err:
            raise OldapErrorValue(f'Invalid version string: "{versionstring}": {err}')

    def increment_patch(self):
        self.__patch += 1

    def increment_minor(self):
        self.__minor += 1
        self.__patch = 0

    def increment_major(self):
        self.__major += 1
        self.__minor = 0
        self.__patch = 0

    @property
    def major(self) -> int:
        return self.__major

    @property
    def minor(self) -> int:
        return self.__minor

    @property
    def patch(self) -> int:
        return self.__patch




from enum import unique, Enum


@unique
class PropClassAttr(Enum):
    """
    Enumeration of all the attributes of a property. Please note that the `Oldap:restriction` attribute
    itself is a complex onbject defining restrictions to the property.
    The following attributes are defined:

    - `PropertyClassAttribute.SUBPROPERTY_OF`: The given property is a specialization of the property given here
    - `PropertyClassAttribute.PROPERTY_OF`: The given property is either a `owl:ObjectProperty` or
      a `owl:DataProperty`. These are important to create a correct OWL ontology. `owl:ObjectProperty`'s
      point to a resource, while `owl:DataProperty`'s point to literal value.
    - `PropertyClassAttribute.TO_NODE_IRI`: The given property points the a resource of the given class.
    - `PropertyClassAttribute.DATATYPE`: Th egiven property must be of the given data type (as defined by
      the XML datatypes. See [XsdDatatypes](xsd_datatypes).
    - `PropertyClassAttribute.RESTRICTION`: Restrictions the property and its values must follow
    - `PropertyClassAttribute.NAME`: Name of the property. The literal is a string that may have attached
      a language id.
    - `PropertyClassAttribute.DESCRIPTION`: Description of the property. The literal is a string that may
      have attached a language id.
    """
    SUBPROPERTY_OF = 'rdfs:subPropertyOf'
    PROPERTY_TYPE = 'rdf:type'
    TO_NODE_IRI = 'sh:class'
    DATATYPE = 'sh:datatype'
    NAME = 'sh:name'
    DESCRIPTION = 'sh:description'
    ORDER = 'sh:order'
    MIN_COUNT = 'sh:minCount'  # used also for OWL ontology
    MAX_COUNT = 'sh:maxCount'  # used also for OWL ontology
    LANGUAGE_IN = 'sh:languageIn'
    UNIQUE_LANG = 'sh:uniqueLang'
    IN = 'sh:in'
    MIN_LENGTH = 'sh:minLength'
    MAX_LENGTH = 'sh:maxLength'
    PATTERN = 'sh:pattern'
    MIN_EXCLUSIVE = 'sh:minExclusive'
    MIN_INCLUSIVE = 'sh:minInclusive'
    MAX_EXCLUSIVE = 'sh:maxExclusive'
    MAX_INCLUSIVE = 'sh:maxInclusive'
    LESS_THAN = 'sh:lessThan'
    LESS_THAN_OR_EQUALS = 'sh:lessThanOrEquals'




from enum import unique, Enum


@unique
class PermissionSetAttr(Enum):
    ID = 'permset:id'  # virtual property, no equivalent in RDF
    LABEL = 'rdfs:label'
    COMMENT = 'rdfs:comment'
    GIVES_PERMISSION = 'oldap:givesPermission'
    DEFINED_BY_PROJECT = 'oldap:definedByProject'

    @property
    def toRdf(self):
        return self.value



from dataclasses import dataclass
from datetime import datetime
from functools import partial
from pprint import pprint
from typing import Union, List, Dict, Callable, Self, Any, TypeVar, Sequence
from pystrict import strict

from oldaplib.src.connection import Connection
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.globalconfig import GlobalConfig
from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.helpers.observable_dict import ObservableDict
from oldaplib.src.helpers.observable_set import ObservableSet
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorNotFound, OldapErrorAlreadyExists, OldapErrorInconsistency, OldapErrorUpdateFailed, \
    OldapErrorValue, OldapErrorKey
from oldaplib.src.enums.propertyclassattr import PropClassAttr
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.enums.resourceclassattr import ResClassAttribute
from oldaplib.src.helpers.semantic_version import SemanticVersion
from oldaplib.src.helpers.tools import RdfModifyRes, RdfModifyItem, lprint
from oldaplib.src.enums.xsd_datatypes import XsdDatatypes
from oldaplib.src.dtypes.bnode import BNode
from oldaplib.src.enums.action import Action
from oldaplib.src.project import Project
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.context import Context
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.propertyclass import PropertyClass, Attributes
from oldaplib.src.xsd.xsd_string import Xsd_string

#
# Datatype definitions
#
RC = TypeVar('RC', bound='ResourceClass')
AttributeTypes = Iri | LangString | Xsd_boolean | ObservableDict[Iri, RC | None] | None
ResourceClassAttributesContainer = Dict[ResClassAttribute, AttributeTypes]
Properties = Dict[BNode, Attributes]
SuperclassParam = Iri | str | list[Iri] | tuple[Iri] | set[Iri] | None
AttributeParams = LangString | Xsd_boolean | SuperclassParam

@dataclass
class ResourceClassAttributeChange:
    old_value: Union[AttributeTypes, PropertyClass, Iri, None]
    action: Action
    test_in_use: bool


@dataclass
class ResourceClassPropertyChange:
    old_value: Union[PropertyClass, Iri, None]
    action: Action
    test_in_use: bool


#@strict
class ResourceClass(Model, Notify):
    _graph: Xsd_NCName
    _project: Project
    _sysproject: Project = None
    _owlclass_iri: Iri | None
    _attributes: ResourceClassAttributesContainer
    _properties: dict[Iri, PropertyClass]
    _attr_changeset: dict[ResClassAttribute, ResourceClassAttributeChange]
    _prop_changeset: dict[Iri, ResourceClassPropertyChange]
    __creator: Iri | None
    __created: Xsd_dateTime | None
    __contributor: Iri | None
    __modified: Xsd_dateTime | None
    __version: SemanticVersion
    __from_triplestore: bool

    __datatypes = {
        ResClassAttribute.SUPERCLASS: ObservableDict,
        ResClassAttribute.LABEL: LangString,
        ResClassAttribute.COMMENT: LangString,
        ResClassAttribute.CLOSED: Xsd_boolean
    }

    def assign_superclass(self, superclass: SuperclassParam) -> ObservableDict[Iri, RC | None]:

        def __check(sc: Any):
            scval = Iri(sc)
            sucla = None
            if scval.is_qname and scval.prefix == self._project.projectShortName:
                sucla = ResourceClass.read(self._con, self._project, scval)
            return scval, sucla

        data = ObservableDict()
        if isinstance(superclass, (list, tuple, set)):
            for sc in superclass:
                iri, sucla = __check(sc)
                data[iri] = sucla
        else:
            iri, sucla = __check(superclass)
            data[iri] = sucla
        data.set_on_change(self.__sc_changed)
        return data


    def __init__(self, *,
                 con: IConnection,
                 project: Project,
                 owlclass_iri: Iri | str | None = None,
                 superclass: SuperclassParam = None,
                 label: LangString | str | None = None,
                 comment: LangString | str | None = None,
                 closed: Xsd_boolean | bool | None = None,
                 properties: List[PropertyClass | Iri] | None = None,
                 notifier: Callable[[PropClassAttr], None] | None = None,
                 notify_data: PropClassAttr | None = None):
        Model.__init__(self, con)
        Notify.__init__(self, notifier, notify_data)
        self._attr_changeset = {}
        self._prop_changeset = {}

        if not isinstance(project, Project):
            raise OldapErrorValue('The project parameter must be a Project instance')
        self._project = project
        if self._sysproject is None:
            self._sysproject = Project.read(self._con, Xsd_NCName("oldap"))
        # if owlclass_iri and owlclass_iri != Iri('oldap:Thing', validate=False):
        #     thingiri = Iri('oldap:Thing', validate=False)
        #     thing = ResourceClass.read(self._con, self._project, thingiri)
        #     self._subClassOf[thingiri] = thing
        # else:
        #     self._subClassOf = {}
        context = Context(name=self._con.context_name)
        context[project.projectShortName] = project.namespaceIri
        context.use(project.projectShortName)
        self._graph = project.projectShortName
        self._attributes = {}

        if isinstance(owlclass_iri, Iri):
            self._owlclass_iri = owlclass_iri
        elif owlclass_iri is not None:
            self._owlclass_iri = Iri(owlclass_iri)
        else:
            self._owlclass_iri = None
        if superclass is not None:
            self._attributes[ResClassAttribute.SUPERCLASS] = self.assign_superclass(superclass)
        if label is not None:
            self._attributes[ResClassAttribute.LABEL] = label if isinstance(label, LangString) else LangString(label)
        if comment is not None:
            self._attributes[ResClassAttribute.COMMENT] = comment if isinstance(comment, LangString) else LangString(comment)
        if closed is not None:
            self._attributes[ResClassAttribute.CLOSED] = closed if isinstance(closed, Xsd_boolean) else Xsd_boolean(closed)
        self._properties = {}
        if properties is not None:
            for prop in properties:
                newprop: PropertyClass | Iri | None = None
                if isinstance(prop, Iri):  # Reference to an external, standalone property definition
                    fixed_prop = Iri(str(prop).removesuffix("Shape"))
                    try:
                        newprop = PropertyClass.read(self._con, self._project, fixed_prop)
                    except OldapErrorNotFound as err:
                        newprop = fixed_prop
                elif isinstance(prop, PropertyClass):  # an internal, private property definition
                    if not prop._force_external:
                        prop._internal = owlclass_iri
                    newprop = prop
                else:
                    #newprop = None
                    raise OldapErrorValue(f'Unexpected property type: {type(prop).__name__}')
                if newprop is not None:
                    self._properties[newprop.property_class_iri] = newprop
                    newprop.set_notifier(self.notifier, newprop.property_class_iri)

        for attr in ResClassAttribute:
            prefix, name = attr.value.split(':')
            setattr(ResourceClass, name, property(
                partial(ResourceClass.__get_value, attr=attr),
                partial(ResourceClass.__set_value, attr=attr),
                partial(ResourceClass.__del_value, attr=attr)))

        self.__creator = con.userIri
        self.__created = None
        self.__contributor = con.userIri
        self.__modified = None
        self.__version = SemanticVersion()
        self.__from_triplestore = False

    def __get_value(self: Self, attr: ResClassAttribute) -> AttributeTypes | PropertyClass | Iri | None:
        return self.__getter(attr)

    def __set_value(self: Self, value: AttributeTypes | PropertyClass | Iri, attr: ResClassAttribute) -> None:
        self.__change_setter(attr, value)

    def __del_value(self: Self, attr: ResClassAttribute) -> None:
        self.__deleter(attr)

    def __getter(self, key: ResClassAttribute | Iri) -> AttributeTypes | PropertyClass | Iri:
        if isinstance(key, ResClassAttribute):
            return self._attributes.get(key)
        elif isinstance(key, Iri):
            return self._properties.get(key)
        else:
            return None

    def __change_setter(self, key: ResClassAttribute | Iri, value: AttributeParams | PropertyClass | Iri) -> None:
        if not isinstance(key, (ResClassAttribute, Iri)):
            raise ValueError(f'Invalid key type {type(key)} of key {key}')
        if getattr(value, 'set_notifier', None) is not None:
            value.set_notifier(self.notifier, key)
        if isinstance(key, ResClassAttribute):
            if self._attributes.get(key) is None:  # Attribute not yet set
                self._attr_changeset[key] = ResourceClassAttributeChange(None, Action.CREATE, False)  # TODO: Check if "check_in_use" must be set
            else:
                if self._attr_changeset.get(key) is None:  # Only first change is recorded
                    self._attr_changeset[key] = ResourceClassAttributeChange(self._attributes[key], Action.REPLACE, False)  # TODO: Check if "check_in_use" must be set
                else:
                    self._attr_changeset[key] = ResourceClassAttributeChange(self._attr_changeset[key].old_value, Action.REPLACE, False)  # TODO: Check if "check_in_use" must be set
            if key == ResClassAttribute.SUPERCLASS:
                self._attributes[key] = self.assign_superclass(value)
            else:
                self._attributes[key] = self.__datatypes[key](value)
        elif isinstance(key, Iri):  # Iri
            if self._properties.get(key) is None:  # Property not set -> CREATE action
                self._prop_changeset[key] = ResourceClassPropertyChange(None, Action.CREATE, False)
                if value is None:
                    try:
                        self._properties[key] = PropertyClass.read(self._con, project=self._project, property_class_iri=key)
                    except OldapErrorNotFound as err:
                        self._properties[key] = Iri(key)
                else:
                    value._internal = self._owlclass_iri  # we need to access the private variable here
                    value._property_class_iri = key  # we need to access the private variable here
                    self._properties[key] = value
            else:  # REPLACE action
                if self._prop_changeset.get(key) is None:
                    self._prop_changeset[key] = ResourceClassPropertyChange(self._properties[key], Action.REPLACE, True)
                else:
                    self._prop_changeset[key] = ResourceClassPropertyChange(self._prop_changeset[key].old_value, Action.REPLACE, True)
                if value is None:
                    try:
                        self._properties[key] = PropertyClass.read(self._con, project=self._project, property_class_iri=key)
                    except OldapErrorNotFound as err:
                        self._properties[key] = key
                else:
                    value._internal = self._owlclass_iri  # we need to access the private variable here
                    value._property_class_iri = key  # we need to access the private variable here
                    self._properties[key] = value
        else:
            raise OldapError(f'Invalid key type {type(key).__name__} of key {key}')
        self.notify()

    def __deleter(self, key: ResClassAttribute | Iri) -> None:
        if not isinstance(key, (ResClassAttribute, Iri)):
            raise ValueError(f'Invalid key type {type(key).__name__} of key {key}')
        if isinstance(key, ResClassAttribute):
            if self._attr_changeset.get(key) is None:
                self._attr_changeset[key] = ResourceClassAttributeChange(self._attributes[key], Action.DELETE, False)
            else:
                self._attr_changeset[key] = ResourceClassAttributeChange(self._attr_changeset[key].old_value, Action.DELETE, False)
            del self._attributes[key]
        elif isinstance(key, Iri):
            if self._prop_changeset.get(key) is None:
                self._prop_changeset[key] = ResourceClassPropertyChange(self._properties[key], Action.DELETE, False)
            else:
                self._prop_changeset[key] = ResourceClassPropertyChange(self._prop_changeset[key].old_value, Action.DELETE, False)
            del self._properties[key]
        self.notify()


    def __getitem__(self, key: ResClassAttribute | Iri) -> AttributeTypes | PropertyClass | Iri:
        return self.__getter(key)

    def get(self, key: ResClassAttribute | Iri) -> AttributeTypes | PropertyClass | Iri | None:
        if isinstance(key, ResClassAttribute):
            return self._attributes.get(key)
        elif isinstance(key, Iri):
            return self._properties.get(key)
        else:
            return None

    def __setitem__(self, key: ResClassAttribute | Iri, value: AttributeParams | PropertyClass | Iri) -> None:
        self.__change_setter(key, value)

    def __delitem__(self, key: ResClassAttribute | Iri) -> None:
        self.__deleter(key)

    @property
    def owl_class_iri(self) -> Iri:
        return self._owlclass_iri

    @property
    def version(self) -> SemanticVersion:
        return self.__version

    @property
    def creator(self) -> Iri | None:
        return self.__creator

    @property
    def created(self) -> Xsd_dateTime | None:
        return self.__created

    @property
    def contributor(self) -> Iri | None:
        return self.__contributor

    @property
    def modified(self) -> Xsd_dateTime | None:
        return self.__modified

    @property
    def properties(self) -> dict[Iri, PropertyClass]:
        return self._properties

    def properties_items(self):
        return self._properties.items()

    def attributes_items(self):
        return self._attributes.items()

    def __str__(self):
        blank = ' '
        indent = 2
        s = f'Shape: {self._owlclass_iri}Shape\n'
        s += f'{blank:{indent*1}}Attributes:\n'
        for attr, value in self._attributes.items():
            s += f'{blank:{indent*2}}{attr.value} = {value}\n'
        s += f'{blank:{indent*1}}Properties:\n'
        sorted_properties = sorted(self._properties.items(), key=lambda prop: prop[1].order if prop[1].order is not None else 9999)
        for qname, prop in sorted_properties:
            s += f'{blank:{indent*2}}{qname} = {prop}\n'
        return s

    def changeset_clear(self) -> None:
        for attr, change in self._attr_changeset.items():
            if change.action == Action.MODIFY:
                self._attributes[attr].changeset_clear()
        self._attr_changeset = {}
        for prop, change in self._prop_changeset.items():
            if change.action == Action.MODIFY:
                self._properties[prop].changeset_clear()
        self._prop_changeset = {}

    def notifier(self, what: ResClassAttribute | Iri):
        if isinstance(what, ResClassAttribute):
            self._attr_changeset[what] = ResourceClassAttributeChange(None, Action.MODIFY, True)
        elif isinstance(what, Iri):
            self._prop_changeset[what] = ResourceClassPropertyChange(None, Action.MODIFY, True)
        self.notify()

    def __sc_changed(self, oldval: ObservableDict[Iri, RC]):
        if self._attr_changeset.get(ResClassAttribute.SUPERCLASS) is None:
            self._attr_changeset[ResClassAttribute.SUPERCLASS] = ResourceClassPropertyChange(oldval, Action.MODIFY, True)

    @property
    def in_use(self) -> bool:
        context = Context(name=self._con.context_name)
        query = context.sparql_context
        query += f"""
        SELECT (COUNT(?resinstances) as ?nresinstances)
        WHERE {{
            ?resinstance rdf:type {self._owlclass_iri} .
            FILTER(?resinstances != {self._owlclass_iri}Shape)
        }} LIMIT 2
        """
        jsonobj = self._con.query(query)
        res = QueryProcessor(context, jsonobj)
        if len(res) != 1:
            raise OldapError('Internal Error in "ResourceClass.in_use"')
        for r in res:
            if int(r.nresinstances) > 0:
                return True
            else:
                return False

    @staticmethod
    def __query_shacl(con: IConnection,
                      project: Project,
                      owl_class_iri: Iri) -> Attributes:
        context = Context(name=con.context_name)
        context[project.projectShortName] = project.namespaceIri
        context.use(project.projectShortName)
        graph = project.projectShortName

        query = context.sparql_context
        query += f"""
        SELECT ?attriri ?value
        FROM {graph}:shacl
        WHERE {{
            BIND({owl_class_iri}Shape AS ?shape)
            ?shape ?attriri ?value
        }}
         """
        jsonobj = con.query(query)
        res = QueryProcessor(context, jsonobj)
        if len(res) == 0:
            raise OldapErrorNotFound(f'Resource with iri "{owl_class_iri}" does not exist."')
        attributes: Attributes = {}
        for r in res:
            attriri = r['attriri']
            if attriri == 'rdf:type':
                tmp_owl_class_iri = r['value']
                if tmp_owl_class_iri == 'sh:NodeShape':
                    continue
                if tmp_owl_class_iri != owl_class_iri:
                    raise OldapError(f'Inconsistent Shape for "{owl_class_iri}": rdf:type="{tmp_owl_class_iri}"')
            elif attriri == 'sh:property':
                continue  # processes later – points to a BNode containing
            else:
                attriri = r['attriri']
                if isinstance(r['value'], Iri):
                    if attributes.get(attriri) is None:
                        attributes[attriri] = []
                    attributes[attriri].append(r['value'])
                elif isinstance(r['value'], Xsd_string):
                    if attributes.get(attriri) is None:
                        attributes[attriri] = []
                    attributes[attriri].append(str(r['value']))
                elif isinstance(r['value'], BNode):
                    pass
                else:
                    if attributes.get(attriri) is None:
                        attributes[attriri] = []
                    attributes[attriri].append(r['value'])
        return attributes

    def _parse_shacl(self, attributes: Attributes) -> None:
        for key, val in attributes.items():
            if key == 'sh:targetClass':
                continue
            if key == 'dcterms:hasVersion':
                self.__version = SemanticVersion.fromString(val[0])
            elif key == 'dcterms:creator':
                self.__creator = val[0]
            elif key == 'dcterms:created':
                self.__created = val[0]
            elif key == 'dcterms:contributor':
                self.__contributor = val[0]
            elif key == 'dcterms:modified':
                self.__modified = val[0]
            elif key == 'sh:node':
                #
                # we expect sh:node only if the superclass is also defined as SHACL and we can read it's
                # definitions. All other superlcasses (referencing external ontologies) are only
                # used in the OWL definitions
                #
                if self._attributes.get(ResClassAttribute.SUPERCLASS) is None:
                    self._attributes[ResClassAttribute.SUPERCLASS] = ObservableDict(on_change=self.__sc_changed)
                if str(val[0]).endswith("Shape"):
                    owliri = Iri(str(val[0])[:-5], validate=False)
                    if owliri.prefix == 'oldap':
                        conf = GlobalConfig(self._con)
                        sysproj = conf.sysproject
                        superclass = ResourceClass.read(self._con, sysproj, owliri)
                    else:
                        superclass = ResourceClass.read(self._con, self._project, owliri)
                    self._attributes[ResClassAttribute.SUPERCLASS][owliri] = superclass
                else:
                    raise OldapErrorInconsistency(f'Value "{val[0]}" must end with "Shape".')
            else:
                attr = ResClassAttribute(key)
                if Iri == self.__datatypes[attr]:
                    self._attributes[attr] = val[0]  # is already QName or AnyIRI from preprocessing
                elif XsdDatatypes == self.__datatypes[attr]:
                    self._attributes[attr] = XsdDatatypes(str(val[0]))
                elif LangString == self.__datatypes[attr]:
                    self._attributes[attr] = LangString(val)
                elif Xsd_boolean == self.__datatypes[attr]:
                    self._attributes[attr] = Xsd_boolean(val[0])
                if getattr(self._attributes[attr], 'set_notifier', None) is not None:
                    self._attributes[attr].set_notifier(self.notifier, attr)

        self.__from_triplestore = True
        self.changeset_clear()

    @staticmethod
    def __query_resource_props(con: IConnection, project: Project, owlclass_iri: Iri) -> List[PropertyClass | Iri]:
        """
        This method queries and returns a list of properties defined in a sh:NodeShape. The properties may be
        given "inline" as BNode or may be a reference to an external sh:PropertyShape. These external shapes will be
        read when the ResourceClass is constructed (see __init__() of ResourceClass).

        :param con: IConnection instance
        :param graph: Name of the graph
        :param owlclass_iri: The QName of the OWL class defining the resource. The "Shape" ending will be added
        :return: List of PropertyClasses/QNames
        """

        context = Context(name=con.context_name)
        context[project.projectShortName] = project.namespaceIri
        context.use(project.projectShortName)
        graph = project.projectShortName

        query = context.sparql_context
        query += f"""
        SELECT ?prop ?attriri ?value ?oo
        FROM {graph}:shacl
        WHERE {{
            BIND({owlclass_iri}Shape AS ?shape)
            ?shape sh:property ?prop .
            OPTIONAL {{
                ?prop ?attriri ?value .
                OPTIONAL {{
                    ?value rdf:rest*/rdf:first ?oo
                }}
            }}
        }}
        """
        jsonobj = con.query(query)
        res = QueryProcessor(context=context, query_result=jsonobj)
        propinfos: Dict[Iri, Attributes] = {}
        #
        # first we run over all triples to gather the information about the properties of the possible
        # BNode based sh:property-Shapes.
        # NOTE: some of the nodes may actually be QNames referencing shapes defines as "standalone" sh:PropertyShape's.
        #
        for r in res:
            if isinstance(r['value'], Iri) and r['value'] == 'rdf:type':
                continue
            if not isinstance(r['attriri'], Iri):
                raise OldapError(f"There is some inconsistency in this shape! ({r['attriri']})")
            propnode = r['prop']  # usually a BNode, but may be a reference to a standalone sh:PropertyShape definition
            prop: PropertyClass | Iri
            if isinstance(propnode, Iri):
                qname = propnode
                propinfos[qname] = propnode
            elif isinstance(propnode, BNode):
                if propinfos.get(propnode) is None:
                    propinfos[propnode]: Attributes = {}
                PropertyClass.process_triple(r, propinfos[propnode])
            else:
                raise OldapError(f'Unexpected type for propnode in SHACL. Type = "{type(propnode)}".')
        #
        # now we collected all the information from the triple store. Let's process the information into
        # a list of full PropertyClasses or QName's to external definitions
        #
        proplist: List[Iri | PropertyClass] = []
        for prop_iri, attributes in propinfos.items():
            if isinstance(attributes, Iri):
                proplist.append(prop_iri)
            else:
                prop = PropertyClass(con=con, project=project)
                prop.parse_shacl(attributes=attributes)
                prop.read_owl()
                if prop._internal != owlclass_iri:
                    OldapErrorInconsistency(f'ERRROR ERROR ERROR')
                proplist.append(prop)
        #prop.set_notifier(self.notifier, prop_iri)
        return proplist

    def __read_owl(self):
        context = Context(name=self._con.context_name)
        query1 = context.sparql_context
        query1 += f"""
        SELECT ?prop ?p ?o
        FROM {self._owlclass_iri.prefix}:onto
        WHERE {{
   	        {self._owlclass_iri.toRdf} rdfs:subClassOf ?prop .
            ?prop ?p ?o .
            FILTER(?o != owl:Restriction)
            FILTER NOT EXISTS {{ ?prop a owl:Class }} .
        }}
        """
        jsonobj = self._con.query(query1)
        res = QueryProcessor(context=context, query_result=jsonobj)
        propdict = {}
        for r in res:
            bnode_id = str(r['prop'])
            if not propdict.get(bnode_id):
                propdict[bnode_id] = {}
            p = r['p']
            if p == 'owl:onProperty':
                propdict[bnode_id]['property_iri'] = r['o']
            elif p == 'owl:onClass':
                propdict[bnode_id]['to_node_iri'] = r['o']
            elif p == 'owl:minQualifiedCardinality':
                propdict[bnode_id]['min_count'] = r['o']
            elif p == 'owl:maxQualifiedCardinality':
                propdict[bnode_id]['max_count'] = r['o']
            elif p == 'owl:qualifiedCardinality':
                propdict[bnode_id]['min_count'] = r['o']
                propdict[bnode_id]['max_count'] = r['o']
            elif p == 'owl:onDatatype':
                propdict[bnode_id]['datatype'] = r['o']
            else:
                print(f'ERROR ERROR ERROR: Unknown restriction property: "{p}"')
        for bn, pp in propdict.items():
            if pp.get('property_iri') is None:
                OldapError('Invalid restriction node: No property_iri!')
            property_iri = pp['property_iri']
            prop = [x for x in self._properties if x == property_iri]
            if len(prop) != 1:
                raise OldapError(f'Property "{property_iri}" of "{self._owlclass_iri}" from OWL has no SHACL definition!')
            self._properties[prop[0]].read_owl()
        #
        # now get all the subClassOf of other classes
        #
        query2 = context.sparql_context
        query2 += f"""
        SELECT ?superclass ?p ?o
        FROM {self._owlclass_iri.prefix}:onto
        WHERE {{
            {self._owlclass_iri.toRdf} rdfs:subClassOf ?superclass .
            FILTER isIRI(?superclass) 
        }}
        """
        jsonobj = self._con.query(query2)
        res = QueryProcessor(context=context, query_result=jsonobj)
        if not self._attributes.get(ResClassAttribute.SUPERCLASS):
            self._attributes[ResClassAttribute.SUPERCLASS] = ObservableDict(on_change=self.__sc_changed)
        for r in res:
            if r['superclass'] not in self._attributes[ResClassAttribute.SUPERCLASS]:
                self._attributes[ResClassAttribute.SUPERCLASS][r['superclass']] = None

    @classmethod
    def read(cls, con: IConnection, project: Project, owl_class_iri: Iri) -> Self:
        if not isinstance(project, Project):
            raise OldapErrorValue('The project parameter must be a Project instance')

        properties: List[Union[PropertyClass, Iri]] = ResourceClass.__query_resource_props(con=con, project=project, owlclass_iri=owl_class_iri)
        resclass = cls(con=con, project=project, owlclass_iri=owl_class_iri, properties=properties)
        for prop in properties:
            if isinstance(prop, PropertyClass):
                prop.set_notifier(resclass.notifier, prop.property_class_iri)
        attributes = ResourceClass.__query_shacl(con, project=project, owl_class_iri=owl_class_iri)
        resclass._parse_shacl(attributes=attributes)
        resclass.__read_owl()
        resclass.changeset_clear()

        return resclass

    def read_modified_shacl(self, *,
                            context: Context,
                            graph: Xsd_NCName,
                            indent: int = 0, indent_inc: int = 4) -> Union[datetime, None]:
        blank = ''
        sparql = context.sparql_context
        sparql += f"{blank:{indent * indent_inc}}SELECT ?modified\n"
        sparql += f"{blank:{indent * indent_inc}}FROM {graph}:shacl\n"
        sparql += f"{blank:{indent * indent_inc}}WHERE {{\n"
        sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self._owlclass_iri}Shape as ?res)\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}?res dcterms:modified ?modified .\n'
        sparql += f"{blank:{indent * indent_inc}}}}"
        jsonobj = self._con.transaction_query(sparql)
        res = QueryProcessor(context, jsonobj)
        if len(res) != 1:
            return None
        return res[0].get('modified')

    def read_modified_owl(self, *,
                          context: Context,
                          graph: Xsd_NCName,
                          indent: int = 0, indent_inc: int = 4) -> Union[datetime, None]:
        blank = ''
        sparql = context.sparql_context
        sparql += f"{blank:{indent * indent_inc}}SELECT ?modified\n"
        sparql += f"{blank:{indent * indent_inc}}FROM {graph}:onto\n"
        sparql += f"{blank:{indent * indent_inc}}WHERE {{\n"
        sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self._owlclass_iri} as ?res)\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}?res dcterms:modified ?modified .\n'
        sparql += f"{blank:{indent * indent_inc}}}}"
        jsonobj = self._con.transaction_query(sparql)
        res = QueryProcessor(context, jsonobj)
        if len(res) != 1:
            return None
        return res[0].get('modified')

    def create_shacl(self, timestamp: Xsd_dateTime, indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        sparql = ''
        # for iri, p in self._properties.items():
        #     if p.internal is None and not p.from_triplestore:
        #         #sparql += p.create_shacl(timestamp=timestamp)
        #         sparql += f'{blank:{(indent + 2)*indent_inc}}{iri}Shape a sh:PropertyShape ;\n'
        #         sparql += p.property_node_shacl(timestamp=timestamp, indent=3) + " .\n"
        #         sparql += "\n"

        sparql += f'{blank:{(indent + 1)*indent_inc}}{self._owlclass_iri}Shape a sh:NodeShape, {self._owlclass_iri.toRdf}'
        sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}sh:targetClass {self._owlclass_iri.toRdf}'
        sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}dcterms:hasVersion {self.__version.toRdf}'
        self.__created = timestamp
        sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}dcterms:created {timestamp.toRdf}'
        sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}dcterms:creator {self.__creator.toRdf}'
        self.__modified = timestamp
        sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}dcterms:modified {timestamp.toRdf}'
        sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}dcterms:contributor {self.__contributor.toRdf}'
        for attr, value in self._attributes.items():
            if attr == ResClassAttribute.SUPERCLASS:
                #
                # In SHACL, superclasses are only added if we have access to it's SHACL definition, that is,
                # if it's given as ResourceClass instance.
                # Superclasses without SHACL definition will be only added to the OWL file for reasoning.
                #
                scset = [f'{iri.toRdf}Shape' for iri, resclass in value.items() if resclass]
                valstr = ", ".join(scset)
                if valstr:
                    sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}sh:node {valstr}'
            else:
                sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}{attr.value} {value.toRdf}'

        sparql += f' ;\n{blank:{(indent + 2) * indent_inc}}sh:property'
        sparql += f'\n{blank:{(indent + 3) * indent_inc}}['
        sparql += f'\n{blank:{(indent + 4) * indent_inc}}sh:path rdf:type'
        sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}]'

        for iri, p in self._properties.items():
            if p.internal is not None:
                sparql += f' ;\n{blank:{(indent + 2)*indent_inc}}sh:property'
                sparql += f'\n{blank:{(indent + 3)*indent_inc}}[\n'
                sparql += p.property_node_shacl(timestamp=timestamp, indent=4)
                sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}]'
            else:
                sparql += f' ;\n{blank:{(indent + 2)*indent_inc}}sh:property {iri}Shape'
        if len(self._properties) > 0:
            sparql += ' .\n'
        return sparql

    def create_owl(self, timestamp: Xsd_dateTime, indent: int = 0, indent_inc: int = 4):
        blank = ''
        sparql = ''
        for iri, p in self._properties.items():
            if not p.from_triplestore:
                sparql += p.create_owl_part1(timestamp, indent + 2) + '\n'

        sparql += f'{blank:{(indent + 2) * indent_inc}}{self._owlclass_iri} rdf:type owl:Class ;\n'
        sparql += f'{blank:{(indent + 3) * indent_inc}}dcterms:hasVersion {self.__version.toRdf} ;\n'
        sparql += f'{blank:{(indent + 3) * indent_inc}}dcterms:created {timestamp.toRdf} ;\n'
        sparql += f'{blank:{(indent + 3) * indent_inc}}dcterms:creator {self.__creator.toRdf} ;\n'
        sparql += f'{blank:{(indent + 3) * indent_inc}}dcterms:modified {timestamp.toRdf} ;\n'
        sparql += f'{blank:{(indent + 3) * indent_inc}}dcterms:contributor {self.__contributor.toRdf} ;\n'
        if self._attributes.get(ResClassAttribute.SUPERCLASS) is not None:
            sc = {x.toRdf for x in self._attributes[ResClassAttribute.SUPERCLASS].keys()}
            valstr = ", ".join(sc)
            sparql += f'{blank:{(indent + 3)*indent_inc}}rdfs:subClassOf {valstr} ,\n'
        else:
            sparql += f'{blank:{(indent + 3)*indent_inc}}rdfs:subClassOf\n'
        i = 0
        for iri, p in self._properties.items():
            sparql += p.create_owl_part2(indent + 4)
            if i < len(self._properties) - 1:
                sparql += ' ,\n'
            else:
                sparql += ' .\n'
            i += 1
        return sparql

    def set_creation_metadata(self, timestamp: Xsd_dateTime):
        self.__created = timestamp
        self.__creator = self._con.userIri
        self.__modified = timestamp
        self.__contributor = self._con.userIri
        self.__from_triplestore = True


    def create(self, indent: int = 0, indent_inc: int = 4) -> None:
        if self.__from_triplestore:
            raise OldapErrorAlreadyExists(f'Cannot create property that was read from triplestore before (property: {self._owlclass_iri}')
        timestamp = Xsd_dateTime.now()
        blank = ''
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f'{blank:{indent * indent_inc}}INSERT DATA {{\n'

        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH {self._graph}:shacl {{\n'
        sparql += self.create_shacl(timestamp=timestamp)
        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'

        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH {self._graph}:onto {{\n'
        sparql += self.create_owl(timestamp=timestamp)
        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'

        sparql += f'{blank:{indent * indent_inc}}}}\n'
        self._con.transaction_start()
        if self.read_modified_shacl(context=context, graph=self._graph) is not None:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'Object "{self._owlclass_iri}" already exists.')
        self._con.transaction_update(sparql)
        modtime_shacl = self.read_modified_shacl(context=context, graph=self._graph)
        modtime_owl = self.read_modified_owl(context=context, graph=self._graph)
        if modtime_shacl == timestamp and modtime_owl == timestamp:
            self._con.transaction_commit()
            self.set_creation_metadata(timestamp=timestamp)
        else:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed(f'Creating resource "{self._owlclass_iri}" failed.')

    def write_as_trig(self, filename: str, indent: int = 0, indent_inc: int = 4) -> None:
        with open(filename, 'w') as f:
            timestamp = Xsd_dateTime.now()
            blank = ''
            context = Context(name=self._con.context_name)
            f.write(context.turtle_context)

            f.write(f'{blank:{indent * indent_inc}}{self._graph}:shacl {{\n')
            f.write(self.create_shacl(timestamp=timestamp))
            f.write(f'\n{blank:{indent * indent_inc}}}}\n')

            f.write(f'{blank:{indent * indent_inc}}{self._graph}:onto {{\n')
            f.write(self.create_owl(timestamp=timestamp))
            f.write(f'{blank:{indent * indent_inc}}}}\n')

    def __update_shacl(self, timestamp: Xsd_dateTime, indent: int = 0, indent_inc: int = 4) -> str:
        if not self._attr_changeset and not self._prop_changeset:
            return ''
        blank = ''
        sparql_list = []

        #
        # First process attributes
        #
        for item, change in self._attr_changeset.items():
            sparql = f'#\n# Process "{item.value}" with Action "{change.action.value}"\n#\n'
            if item == ResClassAttribute.SUPERCLASS:
                #
                # Superclasses are only added to SHACL if they have been supplied as ResourceClass instance.
                # Then the subclass inherits all property definitions!!
                # Other superclasses where we do not have access to a SHACL definition are only added to
                # OWL in order to allow reasoning.
                #
                if change.old_value:
                    old_set = {iri for iri, data in change.old_value.items() if data}
                else:
                    old_set = set()
                if self._attributes[item]:
                    new_set = {iri for iri, data in self._attributes[item].items() if data}
                else:
                    new_set = set()
                to_be_deleted = old_set - new_set
                to_be_added = new_set - old_set
                if to_be_deleted or to_be_added:
                    sparql += f'WITH {self._graph}:shacl\n'
                    if to_be_deleted:
                        sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                        for ov in to_be_deleted:
                            sparql += f'{blank:{(indent + 1) * indent_inc}}?res sh:node {ov.toRdf}Shape .\n'
                        sparql += f'{blank:{indent * indent_inc}}}}\n'
                    if to_be_added:
                        sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                        for nv in to_be_added:
                            sparql += f'{blank:{(indent + 1) * indent_inc}}?res sh:node {nv.toRdf}Shape .\n'
                        sparql += f'{blank:{indent * indent_inc}}}}\n'
                    sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
                    sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.owl_class_iri.toRdf}Shape as ?res)\n'
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?res dcterms:modified ?modified .\n'
                    sparql += f'{blank:{(indent + 1) * indent_inc}}FILTER(?modified = {self.__modified.toRdf})\n'
                    sparql += f'{blank:{indent * indent_inc}}}}'
                    sparql_list.append(sparql)
            else:
                sparql += RdfModifyRes.shacl(action=change.action,
                                             graph=self._graph,
                                             owlclass_iri=self._owlclass_iri,
                                             ele=RdfModifyItem(item.value,
                                                               change.old_value,
                                                               self._attributes[item]),
                                             last_modified=self.__modified)
                sparql_list.append(sparql)
        #
        # now process properties
        #
        for prop, change in self._prop_changeset.items():
            if change.action in {Action.CREATE, Action.REPLACE}:  # do nothing for Action.MODIFY here
                if self._properties[prop].internal is None:
                    sparql = f'#\n# Process "QName" with action "{change.action.value}"\n#\n'
                    sparql += RdfModifyRes.shacl(action=change.action,
                                                 graph=self._graph,
                                                 owlclass_iri=self._owlclass_iri,
                                                 ele=RdfModifyItem('sh:property',
                                                                   change.old_value,
                                                                   Iri(f'{prop}Shape')),
                                                 last_modified=self.__modified)
                    sparql_list.append(sparql)
            elif change.action == Action.DELETE:
                if change.old_value.internal is not None:
                    change.old_value.delete()
                else:
                    sparql = f'#\n# Process "QName" with action "{change.action.value}"\n#\n'

                    sparql += RdfModifyRes.shacl(action=change.action,
                                                 graph=self._graph,
                                                 owlclass_iri=self._owlclass_iri,
                                                 ele=RdfModifyItem('sh:property',
                                                                   None if change.old_value is None else Iri(f'{change.old_value.property_class_iri}Shape'),
                                                                   Iri(f'{prop}Shape')),
                                                 last_modified=self.__modified)
                    sparql_list.append(sparql)

        #
        # Updating the timestamp and contributor ID
        #
        sparql = f'#\n# Update/add dcterms:contributor\n#\n'
        sparql += RdfModifyRes.shacl(action=Action.REPLACE if self.__contributor else Action.CREATE,
                                     graph=self._graph,
                                     owlclass_iri=self._owlclass_iri,
                                     ele=RdfModifyItem('dcterms:contributor', self.__contributor, self._con.userIri),
                                     last_modified=self.__modified)
        sparql_list.append(sparql)

        sparql = f'#\n# Update/add dcterms:modified\n#\n'
        sparql += RdfModifyRes.shacl(action=Action.REPLACE if self.__modified else Action.CREATE,
                                     graph=self._graph,
                                     owlclass_iri=self._owlclass_iri,
                                     ele=RdfModifyItem('dcterms:modified', self.__modified, timestamp),
                                     last_modified=self.__modified)
        sparql_list.append(sparql)

        sparql = " ;\n".join(sparql_list)
        return sparql

    def __update_owl(self, timestamp: Xsd_dateTime, indent: int = 0, indent_inc: int = 4) -> str:
        if not self._attr_changeset and not self._prop_changeset:
            return ''
        blank = ''
        sparql_list = []
        #
        # Adapt OWL for changing *attributes* (where only rdfs:subClassOf is relevant...)
        #
        for item, change in self._attr_changeset.items():
            #
            # we only need to add rdfs:subClassOf to the ontology – all other attributes are irrelevant
            #
            if item == ResClassAttribute.SUPERCLASS:
                sparql = f'#\n# OWL: Process attribute "{item.value}" with Action "{change.action.value}"\n#\n'
                sparql += f'WITH {self._graph}:onto\n'
                old_set = set(change.old_value) if change.old_value else set()
                new_set = set(self._attributes[item]) if self._attributes[item] else set()
                to_be_deleted = old_set - new_set
                to_be_added = new_set - old_set
                if to_be_deleted:
                    sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                    for ov in to_be_deleted:
                        if isinstance(ov, ResourceClass):
                            sparql += f'{blank:{(indent + 1) * indent_inc}}?res rdfs:subClassOf {ov.owl_class_iri.toRdf} .\n'
                        else:
                            sparql += f'{blank:{(indent + 1) * indent_inc}}?res rdfs:subClassOf {ov.toRdf} .\n'
                    sparql += f'{blank:{indent * indent_inc}}}}\n'
                if to_be_added:
                    sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                    for nv in to_be_added:
                        if isinstance(nv, ResourceClass):
                            sparql += f'{blank:{(indent + 1) * indent_inc}}?res rdfs:subClassOf {nv.owl_class_iri.toRdf} .\n'
                        else:
                            sparql += f'{blank:{(indent + 1) * indent_inc}}?res rdfs:subClassOf {nv.toRdf} .\n'
                    sparql += f'{blank:{indent * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.owl_class_iri.toRdf} as ?res)\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?res dcterms:modified ?modified .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}FILTER(?modified = {self.__modified.toRdf})\n'
                sparql += f'{blank:{indent * indent_inc}}}}'
                sparql_list.append(sparql)

        for prop, change in self._prop_changeset.items():
            sparql = f'#\n# OWL: Process property "{prop}" with Action "{change.action.value}"\n#\n'
            sparql += f'WITH {self._graph}:onto\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?res rdfs:subClassOf ?node .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?node ?p ?v\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?res rdfs:subClassOf\n'
                sparql += self._properties[prop].create_owl_part2(indent=2)
                sparql += '\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.owl_class_iri.toRdf} as ?res)\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{(indent + 2) * indent_inc}}?node owl:onProperty {prop.toRdf}\n'
            sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql_list.append(sparql)

        #
        # Updating the timestamp and contributor ID
        #
        sparql = f'#\n# Update/add dcterms:contributor\n#\n'
        sparql += RdfModifyRes.onto(action=Action.REPLACE if self.__contributor else Action.CREATE,
                                    graph=self._graph,
                                    owlclass_iri=self._owlclass_iri,
                                    ele=RdfModifyItem('dcterms:contributor', self.__contributor, self._con.userIri),
                                    last_modified=self.__modified)
        sparql_list.append(sparql)

        sparql = f'#\n# Update/add dcterms:modified\n#\n'
        sparql += RdfModifyRes.onto(action=Action.REPLACE if self.__modified else Action.CREATE,
                                    graph=self._graph,
                                    owlclass_iri=self._owlclass_iri,
                                    ele=RdfModifyItem('dcterms:modified', self.__modified, timestamp),
                                    last_modified=self.__modified)
        sparql_list.append(sparql)

        sparql = " ;\n".join(sparql_list)
        return sparql

    def update(self) -> None:
        timestamp = Xsd_dateTime.now()
        context = Context(name=self._con.context_name)
        #
        # First we process the changes regarding the properties
        #
        for prop, change in self._prop_changeset.items():
            if change.action == Action.CREATE:
                if self._properties[prop].internal is not None:
                    self._properties[prop].create()

                    # TODO: Add here the OWL rdfs:subClassOf to the owl ontology
            elif change.action == Action.REPLACE:
                if change.old_value.internal is not None:
                    change.old_value.delete()
                if not self._properties[prop].from_triplestore:
                    self._properties[prop].create()
                else:
                    if self._properties[prop].get(PropClassAttr.EXCLUSIVE_FOR) is None:
                        continue  # TODO: replace reference in __update_shacl and __update_owl
                    else:
                        raise OldapErrorInconsistency(f'Property is exclusive – simple reference not allowed')
            elif change.action == Action.MODIFY:
                self._properties[prop].update()
            elif change.action == Action.DELETE:
                if change.old_value.internal is not None:
                    change.old_value.delete()
        sparql = context.sparql_context
        sparql += self.__update_shacl(timestamp=timestamp)
        sparql += ' ;\n'
        sparql += self.__update_owl(timestamp=timestamp)

        try:
            self._con.transaction_start()
        except OldapError as err:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_update(sparql)
        except OldapError as err:
            self._con.transaction_abort()
            raise
        try:
            modtime_shacl = self.read_modified_shacl(context=context, graph=self._graph)
            modtime_owl = self.read_modified_owl(context=context, graph=self._graph)
        except OldapError as err:
            self._con.transaction_abort()
            raise
        if modtime_shacl == timestamp and modtime_owl == timestamp:
            self._con.transaction_commit()
            self.changeset_clear()
            self.__modified = timestamp
            self.__contributor = self._con.userIri
        else:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed(f'Update of {self._owlclass_iri} failed. {modtime_shacl} {modtime_owl} {timestamp}')

    def __delete_shacl(self) -> str:
        sparql = f'#\n# SHALC: Delete "{self._owlclass_iri}" completely\n#\n'
        sparql += f"""
        WITH {self._graph}:shacl
        DELETE {{
            {self._owlclass_iri}Shape ?rattr ?rvalue .
            ?rvalue ?pattr ?pval .
            ?z rdf:first ?head ;
            rdf:rest ?tail .
        }}
        WHERE {{
            {self._owlclass_iri}Shape ?rattr ?rvalue .
            OPTIONAL {{
                ?rvalue ?pattr ?pval .
                OPTIONAL {{
                    ?pval rdf:rest* ?z .
                    ?z rdf:first ?head ;
                    rdf:rest ?tail .
                }}
                FILTER(isBlank(?rvalue))
            }}
        }}
        """
        return sparql

    def __delete_owl(self) -> str:
        sparql = f'#\n# OWL: Delete "{self._owlclass_iri}" completely\n#\n'
        sparql += f"""
        WITH {self._graph}:onto
        DELETE {{
            ?prop ?p ?v
        }}
        WHERE {{
            ?prop rdfs:domain {self._owlclass_iri.toRdf} .
            ?prop ?p ?v
        }} ;
        WITH {self._graph}:onto
        DELETE {{
            ?res ?prop ?value .
            ?value ?pp ?vv .
        }}
        WHERE {{
            BIND({self._owlclass_iri.toRdf} AS ?res)
            ?res ?prop ?value
            OPTIONAL {{
                ?value ?pp ?vv
                FILTER(isBlank(?value))
            }}
        }}
        """
        return sparql

    def delete(self) -> None:
        timestamp = datetime.now()
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += self.__delete_shacl()
        sparql += ' ;\n'
        sparql += self.__delete_owl()
        self._con.transaction_start()
        self._con.transaction_update(sparql)
        sparql = context.sparql_context
        sparql += f"SELECT * FROM {self._graph}:shacl WHERE {{ {self._owlclass_iri}Shape ?p ?v }}"
        jsonobj = self._con.transaction_query(sparql)
        res_shacl = QueryProcessor(context, jsonobj)
        sparql = context.sparql_context
        sparql += f"SELECT * FROM {self._graph}:onto WHERE {{ {self._owlclass_iri.toRdf} ?p ?v }}"
        jsonobj = self._con.transaction_query(sparql)
        res_onto = QueryProcessor(context, jsonobj)
        if len(res_shacl) > 0 or len(res_onto) > 0:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed(f'Could not delete "{self._owlclass_iri}".')
        else:
            self._con.transaction_commit()


if __name__ == '__main__':
    context = Context(name="DEFAULT")
    context['test'] = NamespaceIRI("http://oldap.org/test#")
    context.use('test', 'dcterms')

    connection = Connection(server='http://localhost:7200',
                                 userId="rosenth",
                                 credentials="RioGrande",
                                 repo="oldap",
                                 context_name="DEFAULT")
    properties: list[PropertyClass | Iri] = [
        Iri("test:comment"),
        Iri("test:test"),
    ]
    r1 = ResourceClass(con=connection,
                       graph=Xsd_NCName('test'),
                       owlclass_iri=Iri("test:testMyResMinimal"))
    r1.closed = Xsd_boolean(True)




import re
from datetime import datetime
from typing import Self

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_dateTime(Xsd):
    """
    Implements the XML Schema [xsd:dateTime](https://www.w3.org/TR/xmlschema11-2/#dateTime) datatype
    """
    __value: datetime

    def __init__(self, value: datetime | Self | str | None = None, validate: bool = True):
        """
        Constructor of a Xsd_dateTime instance
        :param value: a Xsd_dateTime instance, Python datetime
        object or a string representation of a datetime in ISO format, or None.
        If the parameter is None or omitted, the current datetime is used.
        :type value: datetime | Self | str | None
        :raises OldapErrorValue: if the parameter cannot be converted to a datetime
        """
        if value is None:
            self.__value = datetime.now()
        elif isinstance(value, Xsd_dateTime):
            self.__value = value.__value
        elif isinstance(value, datetime):
            self.__value = value
        else:
            if validate:
                if re.match(
                        r'^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])T([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)(\.\d+)?(Z|[+-]([01][0-9]|2[0-3]):[0-5][0-9])?$',
                        value) is None:
                    raise OldapErrorValue(f'DateTime "{value}" not a valid ISO 8601.')
            try:
                self.__value = datetime.fromisoformat(value)
            except ValueError as err:
                raise OldapErrorValue(str(err))

    def __str__(self) -> str:
        """
        String representation of the Xsd_dateTime instance in ISO format
        :return: ISO string
        """
        return self.__value.isoformat()

    def __repr__(self) -> str:
        """
        String representation of the Xsd_dateTime instance in ISO format as constructor
        :return: Constructor string representation of the Xsd_dateTime instance in ISO format
        """
        return f'Xsd_dateTime("{self.__value.isoformat()}")'

    def __eq__(self, other: Self | str | None) -> bool:
        """
        Equality test for Xsd_dateTime instance
        :param other: Value to compare instance to
        :return: True or False
        :raises OldapErrorValue: if the parameter cannot be converted to a datetime
        """
        if other is None:
            return False
        if isinstance(other, Xsd_dateTime):
            return self.__value == other.__value
        if isinstance(other, datetime):
            return self.__value == other
        else:
            if re.match(
                    r'^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])T([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)(\.\d+)?(Z|[+-]([01][0-9]|2[0-3]):[0-5][0-9])?$',
                    str(other)) is None:
                raise OldapErrorValue(f'DateTime "{other}" not a valid ISO 8601.')
            other = datetime.fromisoformat(other)
            return self.__value == other

    def _as_dict(self) -> dict:
        """
        Used internall for JSON serialization (@serializer decorator)
        :return: dict
        """
        return {'value': self.__value.isoformat()}

    @property
    def toRdf(self) -> str:
        """
        Converts the Xsd_dateTime instance to a RDF string
        :return: RDF string
        """
        return f'"{self.__value.isoformat()}"^^xsd:dateTime'

    @property
    def value(self) -> datetime:
        """
        Converts the Xsd_dateTime instance to a Python datetime object and returns it
        :return: datetime object
        """
        return self.__value

    @classmethod
    def now(cls) -> Self:
        """
        Return a Xsd_dateTime instance set to now()
        :return: Xsd_dateTime instance
        """
        return cls(datetime.now())



import re
from typing import Self

from pystrict import strict

from oldaplib.src.enums.xsd_datatypes import XsdValidator, XsdDatatypes
from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_string import Xsd_string


#@strict
@serializer
class Xsd_token(Xsd):
    """
    Implements the XML Schema [xsd:token](https://www.w3.org/TR/xmlschema11-2/#token) datatype
    """
    __value: str

    def __init__(self, value: Self | str, validate: bool = True):
        """
        Constructor for the Xsd_token class.
        :param value: A Xml_token instance or a valid token string
        :type value: Xsd_token | str
        :raises OldapErrorValue: If the token is invalid
        """
        if isinstance(value, Xsd_token):
            self.__value = value.__value
        else:
            if validate:
                if not XsdValidator.validate(XsdDatatypes.token, value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:token.')
                if not re.match("^[^\\s]+(\\s[^\\s]+)*$", value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:token.')
                if re.match(".*[\n\r\t].*", value) is not None:
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:token.')
            self.__value = value

    def __str__(self):
        """
        String representation of the Xsd_token instance.
        :return: String representation of the Xsd_token instance.
        :rtype: str
        """
        return self.__value

    def __repr__(self):
        """
        Python constructor string representation of the Xsd_token instance.
        :return: Python constructor string representation of the Xsd_token instance.
        :rtype: str
        """
        return f'Xsd_token("{Xsd_string.escaping(str(self))}")'

    def __eq__(self, other: Self | str | None) -> bool:
        """
        Compares two Xsd_token instances for equality.
        :param other: an other Xsd_token instance or a string
        :return: True or False
        """
        if other is None:
            return False
        if isinstance(other, Xsd_token):
            return self.__value == other.__value
        else:
            return self.__value == str(other)


    def __hash__(self) -> int:
        """
        Hash value of the Xsd_token instance.
        :return: Hash value
        :rtype: int
        """
        return super().__hash__()

    @classmethod
    def fromRdf(cls, value: str) -> Self:
        """
        Create a Xsd_token instance from an xs:token RDF Value (string must be without '^^xsd:token' data indicator)
        :param value:
        :return:
        """
        return cls(Xsd_string.unescaping(value), validate=False)


    def _as_dict(self) -> dict[str, str]:
        """
        Internal method used to serialize to JSON (@serializer decorator)
        :return: dict
        """
        return {'value': self.__value}

    @property
    def toRdf(self) -> str:
        """
        RDF representation of the Xsd_token instance.
        :return: RDF representation of the Xsd_token instance.
        :rtype: str
        """
        return f'"{Xsd_string.escaping(str(self))}"^^xsd:token'

    @property
    def value(self) -> str:
        """
        String representation of the Xsd_token instance.
        :return: String representation of the Xsd_token instance.
        """
        return self.__value



import re
from typing import Self

from pystrict import strict

from oldaplib.src.enums.xsd_datatypes import XsdValidator, XsdDatatypes
from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_hexBinary(Xsd):
    """
    Implementation of the XML Schema [xsd:HexBinary](https://www.w3.org/TR/xmlschema11-2/#hexBinary) datatype
    """
    __value: str

    def __init__(self, value: Self | str, validate: bool = True):
        """
        Constructor of the Xsd_hexBinary class
        :param value: Xsd_hexBinary instance or a valid string
        :type value: Xsd_hexBinary | str
        :raises OldapErrorValue: If the value is not a valid Xsd_hexBinary instance
        """
        if isinstance(value, Xsd_hexBinary):
            self.__value = value.__value
        else:
            if validate:
                if not bool(re.match(r'^[0-9A-Fa-f]*$', value)) or len(value) % 2 != 0:
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:hexBinary.')
                if not XsdValidator.validate(XsdDatatypes.hexBinary, value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:hexBinary.')
            self.__value = value

    def __str__(self):
        """
        String representation of the Xsd_hexBinary instance
        :return: String representation of the Xsd_hexBinary instance
        """
        return self.__value

    def __repr__(self):
        """
        Constructor string representation of the Xsd_hexBinary instance
        :return: Constructor string
        """
        return f'Xsd_hexBinary("{str(self)}")'

    def __eq__(self, other: Self | str | None) -> bool:
        """
        Equality test for Xsd_hexBinary instance
        :param other: Xsd_hexBinary instance or a valid string
        :return: True or False
        :raises OldapErrorValue: If the value is not a valid Xsd_hexBinary instance
        """
        if other is None:
            return False
        if isinstance(other, Xsd_hexBinary):
            return self.__value == other.__value
        else:
            return self.__value == Xsd_hexBinary(str(other))

    def __hash__(self) -> int:
        """
        Hash value of the Xsd_hexBinary instance
        :return: Hash value
        """
        return hash(self.__value)

    def _as_dict(self) -> dict[str, str]:
        """
        Internal method for JSON serialization (@serializer decorator)
        :return: dict
        """
        return {'value': self.__value}

    @property
    def toRdf(self) -> str:
        """
        RDF representation of the Xsd_hexBinary instance
        :return: RDF representation of the Xsd_hexBinary instance
        """
        return f'"{str(self)}"^^xsd:hexBinary'

    @property
    def value(self) -> str:
        """
        String representation of the Xsd_hexBinary instance
        :return: String representation of the Xsd_hexBinary instance
        """
        return self.__value
